from pathlib import Path
print('Running' if __name__ == '__main__' else 'Importing', Path(__file__).resolve())

from package.parse.parser import Parser, ParseError
from typing import Set
from pythonds.basic import Stack
from pythonds.trees import BinaryTree

#turn python string expression into binary tree representation, string can be flexible paranthesis optional.
#add support for mathematical propositions (x>1 -> x>0)
class ParseTree(Parser):
    def start(self):
        return self.equiv()

    
    def lit(self):
        # ADDED THIS start
        prevPos = self.pos
        try:
            rv = self.match('ineq')
            return rv
        except ParseError:
            self.pos = prevPos
        # TILL HERE end
        if self.maybe_keyword('('):
            rv = self.match('equiv')
            self.keyword(')')
            return rv
        prevPos = self.pos
        chars = []
        chars.append(self.char('A-Za-z'))
        while True:
            char = self.maybe_char('A-Za-z')
            if char is None:
                break
            chars.append(char)
        if self.maybe_keyword('+','-','*','/','=','>','<', '≤', '≥'):
            self.pos = prevPos
            return None
        rv = BinaryTree('')
        rv.setRootVal(''.join(chars))
        rv.insertLeft('')
        rv.insertRight('')
        return rv
    
    def neg(self):
        if self.maybe_keyword('¬'):
            l = self.match('lit')
            rv = BinaryTree('')
            rv.setRootVal('¬')
            rv.insertLeft('')
            rv.insertRight(l) #TODO:should this be None or ''?
            return rv
        #return self.match('lit')
        rv = self.match('lit')
        if rv is None:
            rv = self.match('ineq')
        return rv

    def disj(self):
        l = self.match('conj')
        if self.maybe_keyword('∨'):
            r = self.match('disj')
            if r is None:
                return None
            rv = BinaryTree('')
            rv.setRootVal('∨')
            rv.insertLeft(l)
            rv.insertRight(r)
            return rv
        return l

    def conj(self):
        l = self.match('neg')
        if self.maybe_keyword('∧'):
            r = self.match('conj')
            if r is None:
                return None
            rv = BinaryTree('')
            rv.setRootVal('∧')
            rv.insertLeft(l)
            rv.insertRight(r)
            return rv
        return l

    def impl(self):
        l = self.match('disj')
        if self.maybe_keyword('→'):
            r = self.match('impl')
            if r is None:
                return None
            rv = BinaryTree('')
            rv.setRootVal('→')
            rv.insertLeft(l)
            rv.insertRight(r)
            return rv
        return l

    def equiv(self):
        l = self.match('impl')
        if self.maybe_keyword('↔'):
            r = self.match('equiv')
            if r is None:
                return None
            rv = BinaryTree('')
            rv.setRootVal('↔')
            rv.insertLeft(l)
            rv.insertRight(r)
            return rv
        return l

    def ineq(self):
        l = self.match('exp')
        while True:
            op = self.maybe_keyword('>', '<', '=', '≤', '≥')
            if op is None:
                break
            r = self.match('exp')
            rv = BinaryTree('')
            rv.setRootVal(op)
            rv.insertLeft(l)
            rv.insertRight(r)
            return rv
        return l

    def exp(self):
        l = self.match('term')
        while True:
            op = self.maybe_keyword('+', '-')
            if op is None:
                break

            r = self.match('term')
            rv = BinaryTree('')
            if op == '+':
                rv.setRootVal('+')
                rv.insertLeft(l)
                rv.insertRight(r)
                return rv
            else:
                rv.setRootVal('+')
                rv.insertLeft(l)
                rv.insertRight(r)
                return rv
        return l

    def term(self):
        l = self.match('factor')
        while True:
            op = self.maybe_keyword('*', '/')
            if op is None:
                break
            r = self.match('factor')
            rv = BinaryTree('')
            if op == '*':
                rv.setRootVal('*')
                rv.insertLeft(l)
                rv.insertRight(r)
                return rv
            else:
                rv.setRootVal('/')
                rv.insertLeft(l)
                rv.insertRight(r)
                return rv
        return l
    
    def factor(self):
        if self.maybe_keyword('('):
            rv = self.match('exp') #TODO: should this be equiv or exp??
            self.keyword(')')
            return rv
    #     elif self.maybe_keyword('¬'):
    #         rv = self.match('variable')
    #         rv = Not(rv)
    #         return rv
        return self.match('number')

    def number(self):
        chars = []
        chars.append(self.char('A-Za-z0-9'))
        while True:
            char = self.maybe_char('A-Za-z0-9')
            if char is None:
                break
            chars.append(char)
        rv = BinaryTree('')
        rv.setRootVal(''.join(chars))
        rv.insertLeft('')
        rv.insertRight('')
        return rv
    



from pathlib import Path
print('Running' if __name__ == '__main__' else 'Importing', Path(__file__).resolve())
from package.parse.parser import Parser, ParseError
from z3 import *

class PropLength(Parser):
    #given f return a tuple representing (length of f, no. of operators in f, no. of props in f, no. of different props, no. of diff operators, list of seen operators, list of seen props)
    #(len, numOp, numProp, numDiffProp, numDiffOp, seenOP, seenProp)
    def start(self):
        return self.equiv()

    def ineq(self):
        l = self.match('exp')
        while True:
            op = self.maybe_keyword('>', '<', '=', '≤', '≥')
            if op is None:
                break
            r = self.match('exp')
            if op not in r[5] and op not in l[5]:
                l[5].update(r[5])
                l[5].add(op)
                l[6].update(r[6])
                rv = (l[0]+r[0]+1, l[1]+r[1]+1, l[2]+r[2], len(l[6]), len(l[5]), l[5], l[6])
            else: #not a new operator
                l[5].update(r[5])
                l[6].update(r[6])
                rv = (l[0]+r[0]+1, l[1]+r[1]+1, l[2]+r[2], len(l[6]), len(l[5]), l[5], l[6])
            # if op == '>':
            #     rv = l > r
            # elif op == '<':
            #     rv = l < r
            # elif op == '=':
            #     rv = l == r
            # elif op == '≥':
            #     rv = l >= r
            # elif op == '≤':
            #     rv = l <= r
            # return rv
        return l

    def exp(self):
        rv = self.match('term')
        while True:
            op = self.maybe_keyword('+', '-')
            if op is None:
                break

            term = self.match('term')
            if op not in rv[5] and op not in term[5]:
                rv[5].update(term[5])
                rv[5].add(op)
                rv[6].update(term[6])
                rv = (rv[0]+term[0]+1, rv[1]+term[1]+1, rv[2]+term[2], len(rv[6]), len(rv[5]), rv[5], rv[6])
            else:
                rv[5].update(term[5])
                rv[6].update(term[6])
                rv = (rv[0]+term[0]+1, rv[1]+term[1]+1, rv[2]+term[2], len(rv[6]), len(rv[5]), rv[5], rv[6])
            # if op == '+':
            #     rv = rv + term
            # else:
            #     rv = rv - term
        return rv

    def term(self):
        rv = self.match('factor')
        while True:
            op = self.maybe_keyword('*', '/')
            if op is None:
                break
            term = self.match('factor')
            if op not in rv[5] and op not in term[5]:
                rv[5].update(term[5])
                rv[5].add(op)
                rv[6].update(term[6])
                rv = (rv[0]+term[0]+1, rv[1]+term[1]+1, rv[2]+term[2], len(rv[6]), len(rv[5]), rv[5], rv[6])
            else:
                rv[5].update(term[5])
                rv[6].update(term[6])
                rv = (rv[0]+term[0]+1, rv[1]+term[1]+1, rv[2]+term[2], len(rv[6]), len(rv[5]), rv[5], rv[6])
            # if op == '*':
            #     rv = rv + term
            # else:
            #     rv = rv + term
        return rv
    
    def factor(self):
        if self.maybe_keyword('('):
            rv = self.match('exp') #TODO: should this be equiv or exp??
            self.keyword(')')
            return rv
    #     elif self.maybe_keyword('¬'):
    #         rv = self.match('variable')
    #         rv = Not(rv)
    #         return rv
        return self.match('number')

    def number(self):
        chars = []
        try:
            chars.append(self.char('A-Za-z')) #removed 0-9 we don't want numbers as variables!
            while True:
                char = self.maybe_char('A-Za-z0-9')
                if char is None:
                    break
                chars.append(char)
            rv = ''.join(chars)
            seenProp = set()
            seenProp.add(rv)
            return (1, 0, 1, 1, 0, set(), seenProp)
        except ParseError:
            chars.append(self.char('0-9'))
            rv = int(''.join(chars))
            return (1, 0, 0, 0, 0, set(), set())

    def lit(self):
        if self.maybe_keyword('('):
            rv = self.match('equiv')
            self.keyword(')')
            return rv
        prevPos = self.pos
        chars = []
        try:
            chars.append(self.char('A-Za-z')) #TODO: then return None if this is an exception???
        except ParseError:
            return None
        while True:
            char = self.maybe_char('A-Za-z')
            if char is None:
                break
            chars.append(char)
        if self.maybe_keyword('+','-','*','/','=','>','<', '≤', '≥'): #TODO: add leq <= and geq >=
            self.pos = prevPos
            return None
        rv = ''.join(chars)
        seenProp = set()
        seenProp.add(rv)
        return (1, 0, 1, 1, 0, set(), seenProp)
    
    def neg(self):
        if self.maybe_keyword('¬'):
            l = self.match('lit')
            if '¬' not in l[5]:
                #seenOp = set()
                l[5].add('¬')
                rv = (1+l[0], 1+l[1], l[2], len(l[6]), len(l[5]), l[5], l[6])
            else:
                rv = (1+l[0], 1+l[1], l[2], len(l[6]), len(l[5]), l[5], l[6])
            return rv
        #return self.match('lit')
        l = self.match('lit')
        if l is None:
            l = self.match('ineq')
        return l

    def disj(self):
        l = self.match('conj')
        if self.maybe_keyword('∨'):
            r = self.match('disj')
            if r is None:
                return None
            l[5].update(r[5])
            l[5].add('∨')
            l[6].update(r[6])
            rv = (l[0]+r[0]+1, l[1]+r[1]+1, l[2]+r[2], len(l[6]), len(l[5]), l[5], l[6])
            return rv
        return l

    def conj(self):
        l = self.match('neg')
        if self.maybe_keyword('∧'):
            r = self.match('conj')
            if r is None:
                return None
            l[5].update(r[5])
            l[5].add('∧')
            l[6].update(r[6])
            rv = (l[0]+r[0]+1, l[1]+r[1]+1, l[2]+r[2], len(l[6]), len(l[5]), l[5], l[6])
            return rv
        return l

    def impl(self):
        l = self.match('disj')
        if self.maybe_keyword('→'):
            r = self.match('impl')
            if r is None:
                return None
            l[5].update(r[5])
            l[5].add('→')
            l[6].update(r[6])
            rv = (l[0]+r[0]+1, l[1]+r[1]+1, l[2]+r[2], len(l[6]), len(l[5]), l[5], l[6])
            return rv
        return l

    def equiv(self):
        l = self.match('impl')
        if self.maybe_keyword('↔'):
            r = self.match('equiv')
            if r is None:
                return None
            l[5].update(r[5])
            l[5].add('↔')
            l[6].update(r[6])
            rv = (l[0]+r[0]+1, l[1]+r[1]+1, l[2]+r[2], len(l[6]), len(l[5]), l[5], l[6])
            return rv
        return l



from pathlib import Path
print('Running' if __name__ == '__main__' else 'Importing', Path(__file__).resolve())
from package.parse.parser import Parser, ParseError
from z3 import *

class PropParser(Parser):
    def start(self):
        # try:
            rv = self.equiv()
        #     self.assert_end()
        # except ParseError:
        #     rv = self.ineq()
            return rv

    def ineq(self):
        l = self.match('exp')
        while True:
            op = self.maybe_keyword('>', '<', '=', '≤', '≥')
            if op is None:
                break
            r = self.match('exp')
            if op == '>':
                rv = l > r
            elif op == '<':
                rv = l < r
            elif op == '=':
                rv = l == r
            elif op == '≥':
                rv = l >= r
            elif op == '≤':
                rv = l <= r
            return rv
        return l

    def exp(self):
        rv = self.match('term')
        while True:
            op = self.maybe_keyword('+', '-')
            if op is None:
                break

            term = self.match('term')
            if op == '+':
                rv = rv + term #Int(rv) + Int(term)
            else:
                rv = rv - term #Int(rv) - Int(term)
        return rv

    def term(self):
        rv = self.match('factor')
        while True:
            op = self.maybe_keyword('*', '/')
            if op is None:
                break
            term = self.match('factor')
            if op == '*':
                rv = rv * term #Int(rv) * Int(term)
            else:
                rv = rv / term #Int(rv) / Int(term)
        return rv
    
    def factor(self):
        if self.maybe_keyword('('):
            rv = self.match('exp') #TODO: should this be equiv or exp??
            self.keyword(')')
            return rv
    #     elif self.maybe_keyword('¬'):
    #         rv = self.match('variable')
    #         rv = Not(rv)
    #         return rv
        return self.match('number')

    def number(self):
        chars = []
        try:
            chars.append(self.char('A-Za-z')) #removed 0-9 we don't want numbers as variables!
            while True:
                char = self.maybe_char('A-Za-z0-9')
                if char is None:
                    break
                chars.append(char)
            rv = ''.join(chars)
            return Int(rv)
        except ParseError:
            chars.append(self.char('0-9'))
            rv = int(''.join(chars))
            return rv

    def lit(self):
        # ADDED THIS start
        # prevPos = self.pos
        # try:
        #     rv = self.match('ineq')
        #     return rv
        # except ParseError:
        #     self.pos = prevPos
        # TILL HERE end
        prevPos = self.pos
        if self.maybe_keyword('('):
            rv = self.match('equiv')
            self.keyword(')')
            # ADDED THIS start
            if self.maybe_keyword('+','-','*','/','=','>','<', '≤', '≥'): #TODO: add leq <= and geq >=
                self.pos = prevPos
                return None
            # TILL HERE end
            return rv
        prevPos = self.pos #and added this
        chars = []
        try:
            chars.append(self.char('A-Za-z')) #TODO: then return None if this is an exception???
        except ParseError:
            return None
        while True:
            char = self.maybe_char('A-Za-z')
            if char is None:
                break
            chars.append(char)
        if self.maybe_keyword('+','-','*','/','=','>','<', '≤', '≥'): #TODO: add leq <= and geq >=
            self.pos = prevPos
            return None
        rv = ''.join(chars)
        return Bool(rv)
    
    def neg(self):
        if self.maybe_keyword('¬'):
            l = self.match('lit')
            rv = Not(l)
            return rv
        #return self.match('lit')
        l = self.match('lit')
        if l is None:
            l = self.match('ineq')
        return l

    def disj(self):
        l = self.match('conj')
        if self.maybe_keyword('∨'):
            r = self.match('disj')
            if r is None:
                return None
            rv = Or(l, r)
            return rv
        return l

    def conj(self):
        l = self.match('neg')
        if self.maybe_keyword('∧'):
            r = self.match('conj')
            if r is None:
                return None
            rv = And(l, r)
            return rv
        return l

    def impl(self):
        l = self.match('disj')
        if self.maybe_keyword('→'):
            r = self.match('impl')
            if r is None:
                return None
            rv = Implies(l, r)
            return rv
        return l

    def equiv(self):
        l = self.match('impl')
        if self.maybe_keyword('↔'):
            r = self.match('equiv')
            if r is None:
                return None
            rv = l == r
            return rv
        return l



from pathlib import Path
print('Running' if __name__ == '__main__' else 'Importing', Path(__file__).resolve())
import z3
from z3 import *
import string
import click
import ast
from main import FormulaGenLEANsyntax
from FormulaGenerator import FormulaGenerator

#global listOfOperators
# global EQ_OPERATORS
# EQ_OPERATORS = ['=', '>', '<', '≤', '≥']
s = Solver()

import sys
current_module = sys.modules[__name__]


#TODO move the cmd line func out and let it call the class
@click.command()
@click.argument('filepath', type=click.Path(exists=True))
def generateFromFile(filepath):
    """
    Call mainResult with the parameters in the file at the given path.
        
    :param filepath: The path to the file.
    :type filepath: String

    :param listOfOperators: The list of lists of operators
    :type listOfOperators: list
    """
    with open(filepath, 'r', encoding='utf-8-sig') as f:
        formula = f.readline()
        operators = ast.literal_eval(f.readline())
        # print(formula)
        # print(operators)
        formulaGen = FormulaGenerator(str(formula), operators)
        #self.mainResult(str(formula), operators)
        formulaGen.mainResult()
        # try:
        #     mainResult(str(formula), operators)
        # except Exception:
        #     mainResultOne(formula, operators)
        try:
            formulaGen.mainResult()
        except Exception:
            formulaLEANGen = FormulaGenLEANsyntax(str(formula), operators)
            formulaLEANGen.mainResultOne()


# def generateFromStr(strFormula, operators):
#     try:
#         formulaGen = FormulaGenerator(str(strFormula), operators)
#         formulaGen.mainResult()
#     except Exception:
#         formulaLEANGen = FormulaGenLEANsyntax(str(strFormula), operators)
#         formulaLEANGen.mainResultOne()



exp1 = '(declare-const x Int) (assert (=> (> x 1) (> x 0)))'
#mainResult(exp1, [['Implies', '=='], ['>','<','==']])
exp2 = '(declare-const P Bool) (declare-const Q Bool) (assert (= (= P Q) (and (=> P Q) (=> Q P))))'
#mainResult(exp2, [['Implies', '=='], ['And', 'Or'], ['Not']]) #TODO: Not isn't being applied
exp3 = '(declare-const x Int) (declare-const y Int) (assert (= (+ (+ x y) 1) (+ x (+ y 1))))'
# mainResult(exp3, [['==', '!='], ['+', '-']])
# maybe I can make this a string then bring it back to z3? How?

# plus = Function('plus', IntSort(), IntSort(), IntSort())
# minus = Function('minus', IntSort(), IntSort(), IntSort())
exp4 = '(declare-fun plus (Int Int) Int) (declare-fun minus (Int Int) Int) (declare-const x Int) (declare-const y Int) (declare-const z Int) (assert (= (plus x (plus y z)) (plus (plus x y) z)))'
formulaGen = FormulaGenerator(strFormula=exp3, listOfOperators=[['==', '!='], ['+', '-']])
formulaGen.mainResult() #mainResult(exp4, [['plus', 'minus']])
#so transitive thing doesn't happen [['Implies', '=='], ['>','<','=='], ['==', 'Or']] 
# probably because the model is computed first before replacement begins (concretization)

z3_exp = parse_smt2_string(exp1)[0]
x1 = z3_exp.children()[0].children()[0]
one1 = z3_exp.children()[0].children()[1]
#formula = x1 > one1
# print(len(x1.children()))
# print(len(one1.children()))
# print(is_bool(Bool('P')))
# print(is_bool(Or(Bool('P'), Bool('Q'))))
#print(formula)

if __name__ == "__main__":  #TODO: uncomment
    generateFromFile()

# exp2 = Implies(Int('x') > 1, Int('x') > 0)
# v = parse_smt2_string('(declare-const x Int) (assert (=> (> x 1) (> x 0)))')
# x1 = v[0].children()[0].children()[0]
# one1 = v[0].children()[0].children()[1]
# print(x1 > one1)
# print('howdy')
# print(exp2.children())
# print(exp2.decl())
# print(v[0])
# testList = [And, Or, Implies]
# varList = [Bool('P'), Bool('Q')]
# print(testList[0](varList[0], varList[1]))
# exp5 = Not(Bool('P'))
# #print(Bool('R')(None, None))
# #exp6 = And(Bool('P'))
# x = Int('x')
# exp7 = (x > 1) == (x > 0)
# print(checkSat(exp7))
# import operator
# print(operator.gt(2,4))


# import sys
# current_module = sys.modules['z3']
# #print(sys.modules) 
# # z3_exp1 = getattr(current_module, 'Implies')(x > 1, x > 2)
# # z3_exp1 = getattr(current_module, 'gt')(x, 2)
# # print(z3_exp1)
# one = 1
# #print(one.decl())
# ast1 = exp7.children()[0].children()[0]
# print(eq(ast1, Int('x')))
# exp8 = x > 1
# so = Solver()
# #so.add(exp8)
# so.add(x1 > one1)
# print(so.check())


from pathlib import Path
# from project.package import propParser
print('Running' if __name__ == '__main__' else 'Importing', Path(__file__).resolve())
# insert root formula and operators to consider
# then abstract the formula
# then produce all permutations of the abstracted formula by changing operators
# can I use z3 for this by abstracting operators into unknown functions (like f represents and, or) then enumerate all models and reconstruct formula for each model?
from typing import List, Set
import builtins
builtins.Z3_LIB_DIRS = ['C:/Users/dania/AppData/Local/Programs/Python/Python39/Lib/site-packages/z3/lib/libz3.dll']
import z3
from z3 import *
from pythonds.basic import Stack
from pythonds.trees import BinaryTree
import operator
import string
#import propParser
#from .parseTree import ParseTree
from package.parseTree import ParseTree
from package.propParser import PropParser
from package.parseFormula import PropLength
from package.parse.parser import ParseError
import click
import ast


global listOfOperators
global EQ_OPERATORS
EQ_OPERATORS = ['=', '>', '<', '≤', '≥']

@click.command()
@click.argument('filepath', type=click.Path(exists=True))
def main(filepath):
    with open(filepath, 'r', encoding='utf-8-sig') as f:
        formula = f.readline()
        operators = ast.literal_eval(f.readline())
        print(formula)
        print(operators)
        #mainResultOne(formula, operators)
        fGenLean = FormulaGenLEANsyntax(listOfOperators=operators, strFormula=formula)
        fGenLean.mainResultOne()

#take in list of lists of operators
# @click.command()
# @click.argument('strFormula')
# @click.argument('operators', type=List)

class FormulaGenLEANsyntax:
    def __init__(self, listOfOperators, strFormula):
        self.listOfOperators = listOfOperators
        self.strFormula = strFormula
        self.idx = 0
        self.var = string.ascii_lowercase
        self.modelIdx = 0
        self.EQ_OPERATORS = ['=', '>', '<', '≤', '≥']
        self.s = Solver()

    def mainResultOne(self): #,strFormula, operators):
        #click.echo('Enter the formula as a string and a list of lists of operators')
        #global listOfOperators
        #listOfOperators = operators
        parser = ParseTree()
        pt = parser.parse(self.strFormula)
        print(pt)
        self.absTree(pt)
        self.conc(pt, self.s)
        #print(postorder(pt))
        listOfAllOper = []
        for l in self.listOfOperators:
            listOfAllOper.extend(l)
        #print(getModels())
        # result = getFormulas(getModels(), pt)
        formulas = self.getFormulas(self.getModels(), pt)
        result = self.getFormulasFormatted(formulas)
        result.extend(self.parenCombFormatted(self.parenComb(self.strFormula, listOfAllOper)))
        print(result)

# def buildParseTree(fpexp):
#     fplist = fpexp.split()
#     pStack = Stack()
#     eTree = BinaryTree('')
#     pStack.push(eTree)
#     currentTree = eTree

#     for i in fplist:
#         if i == '(':
#             currentTree.insertLeft('')
#             pStack.push(currentTree)
#             currentTree = currentTree.getLeftChild()

#         elif i in ['∧', '∨', '→', '↔']:
#             currentTree.setRootVal(i)
#             currentTree.insertRight('')
#             pStack.push(currentTree)
#             currentTree = currentTree.getRightChild()

#         elif i == ')':
#             currentTree = pStack.pop()

#         elif i not in ['∧', '∨', '→', '↔']:
#             try:
#                 currentTree.setRootVal(i)
#                 parent = pStack.pop()
#                 currentTree = parent

#             except ValueError:
#                 raise ValueError("token '{}' is not a valid integer".format(i))
#     return eTree

    def postorder(self, tree):
        if tree != None:
            self.postorder(tree.getLeftChild())
            self.postorder(tree.getRightChild())
            print(tree.getRootVal())

    def absTree(self,tree):
        if tree != None:
            self.absTree(tree.getLeftChild())
            self.absTree(tree.getRightChild())
            global listOfOperators
            rootVal = tree.getRootVal()
            for index,listOp in enumerate(self.listOfOperators):
                if rootVal in listOp:
                    tree.setRootVal('t' + str(index))
                    break

    def numConc(self, s):
        global listOfOperators
        for index,listOp in enumerate(self.listOfOperators):
            if s == 't' + str(index):
                return len(listOp)
        return 1

    def numOfConcretizations(self, abstractedTree):
        res = 1
        if abstractedTree != None:
            res *= self.numConc(abstractedTree.getRootVal())
        return res

    # var = string.ascii_lowercase
    # global idx 
    # idx = 0
    def freshVariable(self):
        global idx
        res = self.var[self.idx]
        self.idx += 1
        return Int(res)


    
    def conc(self, abstractedTree, solver):
        if abstractedTree != None:
            self.conc(abstractedTree.getLeftChild(), solver)
            global listOfOperators
            rootVal = abstractedTree.getRootVal()
            for index in range(len(self.listOfOperators)):
                if rootVal == 't' + str(index):
                    x = self.freshVariable()
                    numConc = self.numOfConcretizations(abstractedTree)
                    
                    solver.add(0<=x, x<numConc)
                    # if abstractedTree != None:
                    #     conc(abstractedTree.getLeftChild(), solver)
                    #     conc(abstractedTree.getRightChild(), solver)
                    self.conc(abstractedTree.getRightChild(), solver)

    def getModels(self):
        result = []
        while self.s.check() == sat:
            m = self.s.model()
            dict = {}
            for d in m:
                dict[str(d)] = m[d]
            result.append(dict)
            # Create a new constraint the blocks the current model
            block = []
            for d in m:
                # d is a declaration
                if d.arity() > 0:
                    raise Z3Exception("uninterpreted functions are not supported")
                # create a constant from declaration
                c = d()
                if is_array(c) or c.sort().kind() == Z3_UNINTERPRETED_SORT:
                    raise Z3Exception("arrays and uninterpreted sorts are not supported")
                block.append(c != m[d])
            self.s.add(Or(block))
        #print(result)
        return result


    # global modelIdx
    def getFormulas(self, listOfModels, abstractedTree):
        result = []
        #print(listOfModels)
        for l in listOfModels:
            #print(l)
        #for i in range(len(var)):
            #global modelIdx
            self.modelIdx = 0
            list = self.getFormula(l, abstractedTree)
            formula = " ".join(list)
            result.append(formula)
            # parser = PropParser()
            # z3_exp = parser.parse(formula)
            # # print(z3_exp)
            # valid = checkValid(z3_exp)
            # sat = checkSat(z3_exp)
            # #result.append(" ".join(list))
            # result.append(" ".join([formula, valid, sat]))
        return result

    def getFormulasFormatted(self, flist):
        result = []
        diffList = self.listDiff(flist, 0, 1)
        for index,formula in enumerate(flist):
            parser = PropParser()
            z3_exp = parser.parse(formula)
            valid = self.checkValid(z3_exp)
            sat = self.checkSat(z3_exp)
            diff = diffList[index]
            result.append(" ".join([formula, valid, sat, str(diff)]))
        return result



    def getFormula(self, model, abstractedTree): 
        #print(model) 
        formula = []
        #i = 0
        #global modelIdx
        x = self.var[self.modelIdx] #x = var[i]
        # print("x: ")
        # print(x)
        # print("model: ")
        # print(model[Int(x)])
        if abstractedTree != None:
            formula.extend(self.getFormula(model, abstractedTree.getLeftChild()))
            rootVal = abstractedTree.getRootVal()
            global listOfOperators
            matched = False
            for index,listOp in enumerate(self.listOfOperators):
                if rootVal == 't' + str(index):
                    # print(Int(var[modelIdx]))
                    # idx = model[Int(var[modelIdx])].as_long()
                    idx = model[(self.var[self.modelIdx])].as_long()
                    formula.append(listOp[idx])
                    self.modelIdx += 1
                    matched = True
            if not matched:
                formula.append(abstractedTree.getRootVal())
            formula.extend(self.getFormula(model, abstractedTree.getRightChild()))
        return formula

    def getFormula(self, model, abstractedTree):  
        formula = []
        #global modelIdx
        x = self.var[self.modelIdx]
        if abstractedTree != None:
            formula.extend(self.getFormula(model, abstractedTree.getLeftChild()))
            rootVal = abstractedTree.getRootVal()
            #global listOfOperators
            matched = False
            for index,listOp in enumerate(self.listOfOperators):
                if rootVal == 't' + str(index):
                    # print(Int(var[modelIdx]))
                    # idx = model[Int(var[modelIdx])].as_long()
                    idx = model[(self.var[self.modelIdx])].as_long()
                    formula.append(listOp[idx])
                    self.modelIdx += 1
                    matched = True
            if not matched:
                formula.append(abstractedTree.getRootVal())
            formula.extend(self.getFormula(model, abstractedTree.getRightChild()))
        return formula

    def checkSat(self, z3_exp):
        solver = Solver()
        solver.add(z3_exp)
        return str(solver.check())

    def checkValid(self, z3_exp):
        solver = Solver()
        solver.add(Not(z3_exp))
        if solver.check() == unsat:
            return 'valid'
        return 'invalid'


# def allParenPlacement(f, operators):
#     #first remove existing brackets
#     f.replace(')', '')
#     f.replace('(', '')
#     n = 0
#     for o in operators:
#         n += f.count(o)
#     n = n/2
#     combList = listParens([], 0, n, 0, 0, [])

# def listParens(l, pos, n, open, close, res):
#     if(close == n):
#         # l = []
#         # for i in l:
#         #     res.append(i)
#         # res.append(l)
#         # return res
#         for i in l:
#             print(i, end = "")
#         print()
#         return
#     else:
#         if(open > close):
#             l[pos] = ')'
#             listParens(l, pos + 1, n, open, close + 1, res)
#         if(open < n):
#             l[pos] = '('
#             listParens(l, pos + 1, n, open + 1, close, res)


    def parenComb(self, formula, allOperators):
        f = formula.replace(')', '')
        f = f.replace('(', '')
        listf = f.split()
        res = []
        global EQ_OPERATORS
        for i in range(0, len(f)):
            if f[i] not in allOperators and f[i] != ' ' and f[i] != ')':
                for j in range(i+1, len(f)):
                    if f[j] in EQ_OPERATORS:
                            break
                    if f[j] not in allOperators and f[j] != ' ' and f[j] != '(':
                        str = f[:i] + '(' + f[i:j+1] + ')' + f[j+1:]
                        res.append(str)
        return res

    def parenCombFormatted(self, flist):
        result = []
        parser = PropParser()
        for f in flist:
            try:
                z3_exp = parser.parse(f)
            except ParseError:
                print(f)
                break
            #print(z3_exp)
            valid = self.checkValid(z3_exp)
            sat = self.checkSat(z3_exp)
            result.append(" ".join([f, valid, sat]))
        return result

    def syntacticDifficulty(self, formula):
        parser = PropLength() #TODO: What to consider? # of props, # of operations, 
        diffTuple = parser.parse(formula)
        #(len, numOp, numProp, numDiffProp, numDiffOp, seenOP, seenProp)
        res = 0.2*diffTuple[0] + 0.1*diffTuple[1] + 0.1*diffTuple[2] + 0.3*diffTuple[3] + 0.3*diffTuple[4]
        return res

    def listDiff(self, flist, minRange, maxRange):
        diffs = []
        res = []
        for f in flist:
            diffs.append(self.syntacticDifficulty(f))
        maxD = max(diffs)
        minD = 0
        for d in diffs:
            r = (maxRange - minRange) * ((d - minD)/(maxD - minD)) + minRange
            res.append(r)
        return res



#print(parenComb('5 * 3 / 2 - 4', ['*', '/', '-']))
#print(listParens([''] * 2 * 2, 0, 2, 0, 0, []))
#     strList = str.split()
#     v1 = Bool('v1')
#     v2 = Bool('v2')
#     for s in strList:
#         if s == '∨':
#             v1 = Or(v1, v2)
#         elif s == '∧':
#             v1 = And(v1, v2)
#         elif s == '→'

#abs('(P ∨ P) ↔ P')

# if __name__ == "__main__":  #TODO: uncomment
#     main()

# print(z3.__file__)

fGenLean = FormulaGenLEANsyntax(listOfOperators=[['∨', '∧'], ['→', '↔']], strFormula='P ∨ P ↔ P')
fGenLean.mainResultOne()

# S = set(('P'))
# print(S)
# S.add('U')
# print(S)
# S1 = set(('P'))
# S2 = set(('Q', 'P'))
# S1.update(S2)
# print(S1)
# S = set()
# S.add('P')
# print(S)
# parser = PropLength() #TODO: What to consider? # of props, # of operations, 
# p = parser.parse('P ∨ ¬Q ∨ R ↔ P') #(8, 4, 4, 3, 3, {'∨', '¬', '↔'}, {'R', 'P', 'Q'})
# print(p)
# print(syntacticDifficulty('P ∨ ¬Q ∨ R ↔ P'))
#(len, numOp, numProp, numDiffProp, numDiffOp, seenOP, seenProp)

#main(r"C:\Users\dania\Documents\college\Fall2021Co-op\fall-2021\test.txt")
#mainResult('(P ↔ Q) ↔ ((P → Q) ∧ (Q → P))', [['∨', '∧'], ['→', '↔'], ['¬']])
#mainResult('(x + y) + 1 = x + (y + 1)', [['+', '-'], ['=']]) #TODO: Parse Error
#mainResult('P ∨ ¬Q ∨ P ↔ P', [['∨', '∧'], ['→', '↔'], ['¬']])
#mainResult('P ∨ P ↔ P', [['∨', '∧'], ['→', '↔']])
#mainResult('x > 1 → x > 0', [['>', '<'], ['→', '↔']])
# sf = Solver()
# exp = (Int('x') < 1) == (Int('x') > 0)
# exp2 = Implies(Int('x') > 1, Int('x') > 0)
# exp3 = Not(Implies(Int('x') > 1, Int('x') > 0))
# parser = PropParser()
# exp4 = parser.parse('x > 1 → x > 0')
# print(checkSat(exp))
# print(checkValid(exp))
# print(exp2)
# print(checkValid(exp2))
# print(checkSat(exp3))
# print(exp4)
# print(checkValid(exp4))
# print(prove(exp2 == exp4))

# sf.add(exp)
# print(sf.check())
# print(exp)

# parser = ParseTree()
# pt = parser.parse('P ∨ P ↔ P')
# print(pt)
# #pt = buildParseTree("( ( P ∨ P ) ↔ P )")
# #pt = buildParseTree("( ( ( P → P ) → P ) → P )") 
# #pt = buildParseTree("( ( ( P → Q ) → P ) → ( x > 0 ) )") 
# pt.postorder() 
# absTree(pt)
# pt.postorder()

# conc(pt, s)
# print(s)
# print(s.check())
# print(numOfConcretizations(pt))
# #print(getModels())
# print(getFormulas(getModels(), pt))
# #print(getModels().sort())



from pathlib import Path
print('Running' if __name__ == '__main__' else 'Importing', Path(__file__).resolve())

#precedence: The ¬ operator has higher precedence than ∧; ∧ has higher precedence than ∨; and ∨ has higher precedence than ⇒ and ⇔.
#code adapted from https://www.booleanworld.com/building-recursive-descent-parsers-definitive-guide/

class ParseError(Exception):
    def __init__(self, pos, msg, *args):
        self.pos = pos
        self.msg = msg
        self.args = args

    def __str__(self):
        return '%s at position %s' % (self.msg % self.args, self.pos)

class Parser:
    def __init__(self):
        self.cache = {}

    def parse(self, text):
        self.text = text
        self.pos = -1
        self.len = len(text) - 1
        rv = self.start()
        self.assert_end()
        return rv

    def assert_end(self):
        if self.pos < self.len:
            raise ParseError(
                self.pos + 1,
                'Expected end of string but got %s',
                self.text[self.pos + 1]
            )
    
    def eat_whitespace(self):
        while self.pos < self.len and self.text[self.pos + 1] in " \f\v\r\t\n":
            self.pos += 1

    def split_char_ranges(self, chars):
        try:
            return self.cache[chars]
        except KeyError:
            pass

        rv = []
        index = 0
        length = len(chars)

        while index < length:
            if index + 2 < length and chars[index + 1] == '-':
                if chars[index] >= chars[index + 2]:
                    raise ValueError('Bad character range')

                rv.append(chars[index:index + 3])
                index += 3
            else:
                rv.append(chars[index])
                index += 1

        self.cache[chars] = rv
        return rv

    def char(self, chars=None):
        if self.pos >= self.len:
            raise ParseError(
                self.pos + 1,
                'Expected %s but got end of string',
                'character' if chars is None else '[%s]' % chars
            )

        next_char = self.text[self.pos + 1]
        if chars == None:
            self.pos += 1
            return next_char

        for char_range in self.split_char_ranges(chars):
            if len(char_range) == 1:
                if next_char == char_range:
                    self.pos += 1
                    return next_char
            elif char_range[0] <= next_char <= char_range[2]:
                self.pos += 1
                return next_char

        raise ParseError(
            self.pos + 1,
            'Expected %s but got %s',
            'character' if chars is None else '[%s]' % chars,
            next_char
        )

    def keyword(self, *keywords):
        self.eat_whitespace()
        if self.pos >= self.len:
            raise ParseError(
                self.pos + 1,
                'Expected %s but got end of string',
                ','.join(keywords)
            )

        for keyword in keywords:
            low = self.pos + 1
            high = low + len(keyword)

            if self.text[low:high] == keyword:
                self.pos += len(keyword)
                self.eat_whitespace()
                return keyword

        raise ParseError(
            self.pos + 1,
            'Expected %s but got %s',
            ','.join(keywords),
            self.text[self.pos + 1],
        )

    def match(self, *rules):
        self.eat_whitespace()
        last_error_pos = -1
        last_exception = None
        last_error_rules = []

        for rule in rules:
            initial_pos = self.pos
            try:
                rv = getattr(self, rule)()
                self.eat_whitespace()
                return rv
            except ParseError as e:
                self.pos = initial_pos

                if e.pos > last_error_pos:
                    last_exception = e
                    last_error_pos = e.pos
                    last_error_rules.clear()
                    last_error_rules.append(rule)
                elif e.pos == last_error_pos:
                    last_error_rules.append(rule)

        if len(last_error_rules) == 1:
            raise last_exception
        else:
            raise ParseError(
                last_error_pos,
                'Expected %s but got %s',
                ','.join(last_error_rules),
                self.text[last_error_pos]
            )
    
    def maybe_char(self, chars=None):
        try:
            return self.char(chars)
        except ParseError:
            return None

    def maybe_match(self, *rules):
        try:
            return self.match(*rules)
        except ParseError:
            return None

    def maybe_keyword(self, *keywords):
        try:
            return self.keyword(*keywords)
        except ParseError:
            return None

    



from setuptools import find_packages, setup

from codecs import open
from os import path

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='formula-generator',
    packages=find_packages(include=['project', 'project.package', 'project.package.parse']),
    include_package_data=True,
    version='0.1.0',
    description='My first Python library',
    url="https://formula-generator.readthedocs.io/",
    author='Dania',
    license='MIT',
    install_requires=['click', 'pythonds', 'z3'],
)


import z3
from z3 import *
import string

s = Solver()

import sys
current_module = sys.modules[__name__]

class FormulaGenerator:
    """
    Generate a list of formulas from the given formula.
    Operators in the formula will be substitited by operators in same lists from the given list of lists of operators.
    
    :param strFormula: The base formula.
    :type strFormula: String
    """
    def __init__(self, listOfOperators, strFormula):
        self.listOfOperators = listOfOperators
        self.strFormula = strFormula
        self.idx = 0
        self.var = string.ascii_lowercase
        self.modelIdx = 0


    def parseListofListOp(self, lop):
        """
        Parse the list of lists of operators into z3 format, return the parsed list.
        
        :param lop: The list of lists of operators to parse.
        :type lop: list
        """
        result = []
        for list in lop:
            result.append(self.parseListOp(list))
        return result

    def parseListOp(self, l):
        """
        Parse a list of operators into z3 format
        
        :param l: The list of operators to parse.
        :type l: list
        """
        result = []
        for op in l:
            if op == '=>' or op == '→' or op.lower() == 'implies':
                result.append('Implies')
            elif op == '<=>' or op == '↔' or op.lower() == 'equiv':
                result.append('==')
            elif op == '∧' or op.lower() == 'and':
                result.append('And')
            elif op == '∨' or op.lower() == 'or':
                result.append('Or')
            elif op == '¬' or op.lower() == 'not':
                result.append('Not')
            elif op == '=' or op.lower() == 'equal':
                result.append('==')
            else:
                result.append(op)
        

    # strFormula in SMT 2.0 format
    def mainResult(self): #, strFormula, operators):
        """
        Return a list of formulas similar to the given formula
        
        :param strFormula: The base formula used to generate similar formulas.
        :type strFormula: String

        :param operators: The list of lists of operators to replace. If two operateros are in the same list then they are replacable with each other.
        :type operators: list
        """
        #global listOfOperators
        #listOfOperators = operators
        z3_exp = parse_smt2_string(self.strFormula)[0]
        self.conc(z3_exp, s)
        listOfAllOper = []
        for l in self.listOfOperators:
            listOfAllOper.extend(l)
        formulas = self.getFormulas(self.getModels(), z3_exp)
        result = self.getFormulasFormatted(formulas)
        #result.extend(parenCombFormatted(parenComb(strFormula, listOfAllOper)))
        print(result)
        #what does children() return when z3 exp has no children? 

    def numConc(self, s):
        #global listOfOperators
        for index,listOp in enumerate(self.listOfOperators):
            if str(s) in listOp:
                return len(listOp)
        return 1

    
    def numOfConcretizations(self, z3_exp):
        res = 1
        if z3_exp != None:
            res *= self.numConc(z3_exp.decl())
        return res

    #var = string.ascii_lowercase
    # global idx 
    # idx = 0
    def freshVariable(self):
        # global idx
        res = self.var[self.idx]
        self.idx += 1
        return Int(res)

    def getRightChild(self, z3_exp):
        children = z3_exp.children()
        if len(children) == 0:
            return None
        elif len(children) == 1:
            return None
        else:
            return children[1]

    def getLeftChild(self, z3_exp):
        children = z3_exp.children()
        if len(children) == 0:
            return None
        else:
            return children[0]
    
    def conc(self, z3_exp, solver):
        if z3_exp != None:
            self.conc(self.getLeftChild(z3_exp), solver)
            #global listOfOperators
            rootVal = z3_exp.decl()
            for l in self.listOfOperators:
                if str(rootVal) in l:
                    x = self.freshVariable()
                    numConc = self.numOfConcretizations(z3_exp)
                    
                    solver.add(0<=x, x<numConc)
                    # if abstractedTree != None:
                    #     conc(abstractedTree.getLeftChild(), solver)
                    #     conc(abstractedTree.getRightChild(), solver)
                    self.conc(self.getRightChild(z3_exp), solver)

    def getModels(self):
        result = []
        while s.check() == sat:
            m = s.model()
            dict = {}
            for d in m:
                dict[str(d)] = m[d]
            result.append(dict)
            # Create a new constraint the blocks the current model
            block = []
            for d in m:
                # d is a declaration
                if d.arity() > 0:
                    raise Z3Exception("uninterpreted functions are not supported")
                # create a constant from declaration
                c = d()
                if is_array(c) or c.sort().kind() == Z3_UNINTERPRETED_SORT:
                    raise Z3Exception("arrays and uninterpreted sorts are not supported")
                block.append(c != m[d])
            s.add(Or(block))
        #print(result)
        return result

    #global modelIdx
    def getFormulas(self, listOfModels, z3_exp):
        result = []
        for l in listOfModels:
            #global modelIdx
            self.modelIdx = 0
            # list = getFormula(l, z3_exp)
            # formula = " ".join(list)
            formula = self.getFormula(l, z3_exp)
            if formula != None:
                result.append(formula)
        return result

    def getFormula(self, model, z3_exp): 
        #global modelIdx
        #x = var[modelIdx]
        formula = None
        if z3_exp != None:
            leftF = self.getFormula(model, self.getLeftChild(z3_exp))
            rootVal = z3_exp.decl()
            if str(rootVal) == 'Int' or len(z3_exp.children()) == 0: # or is_int(rootVal) or is_real(rootVal) or is_bool(rootVal):
                rootVal = z3_exp
            #global listOfOperators
            matched = False
            for listOp in self.listOfOperators:
                if str(rootVal) in listOp:
                    idx = model[(self.var[self.modelIdx])].as_long()
                    rv = listOp[idx]
                    self.modelIdx += 1
                    matched = True
            rightF = self.getFormula(model, self.getRightChild(z3_exp))
            
            if leftF is not None:
                if not matched:
                    formula = self.joinFormulasZ3(str(rootVal), leftF, rightF) #getattr(current_module, rootVal)(leftF, rightF)
                else:
                    formula = self.joinFormulasZ3(str(rv), leftF, rightF) #getattr(current_module, rv)(leftF, rightF)
            else:
                if not matched:
                    formula = rootVal
                else:
                    formula = rv
        return formula

    def joinFormulasZ3(self, op, leftF, rightF):
        formula = None
        try:
            formula = getattr(current_module, op)(leftF, rightF)
        except AttributeError:
            if op == '+':
                formula = leftF + rightF
            elif op == '-':
                formula = leftF - rightF
            elif op == '*':
                formula = leftF * rightF
            elif op == '/':
                formula = leftF / rightF
            elif op == '==':
                formula = leftF == rightF
            elif op == '<':
                formula = leftF < rightF
            elif op == '>':
                formula = leftF > rightF
            elif op == '>=':
                formula = leftF >= rightF
            elif op == '<=':
                formula = leftF <= rightF
            elif op == '!=':
                formula = leftF != rightF
            # elif op == '=':
            #     formula = leftF = rightF
            elif op == '**':
                formula = leftF ** rightF
            elif op == '%':
                formula = leftF % rightF
        return formula


    def getFormulasFormatted(self, flist):
        result = []
        #diffList = listDiff(flist, 0, 1)
        for index,z3_exp in enumerate(flist):
            valid = self.checkValid(z3_exp)
            sat = self.checkSat(z3_exp)
            #diff = diffList[index]
            result.append(" ".join([str(z3_exp), valid, sat])) #TODO:, str(diff)]))
        return result

    def checkSat(self, z3_exp):
        solver = Solver()
        solver.add(z3_exp)
        return str(solver.check())

    def checkValid(self, z3_exp):
        solver = Solver()
        solver.add(Not(z3_exp))
        if solver.check() == unsat:
            return 'valid'
        return 'invalid'
